##
# \file admm_linear_solver_parameter_study.py
# \brief Class to run parameter study for ADMMLinearSolver
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       Aug 2017
#

import numpy as np

import pysitk.python_helper as ph
from nsol.solver_parameter_study import SolverParameterStudy
import nsol.admm_linear_solver as admm


class ADMMLinearSolverParameterStudy(SolverParameterStudy):

    ##
    # Store information for parameter study
    # \date       2017-09-13 23:16:59+0100
    #
    # \param      self                 The object
    # \param      solver               Solver object
    # \param      observer             Observer object
    # \param      dir_output           Output directory to write study results
    # \param      name                 Name of study, string (no white spaces)
    # \param      parameters           Dictionary holding information on
    #                                  varying parameter to sweep through
    #                                  during parameter study
    # \param      reconstruction_info  Dictionary holding information useful
    #                                  for later reconstruction, e.g. image
    #                                  space or sitk.Image to include image
    #                                  header information
    #
    def __init__(self,
                 solver,
                 observer,
                 dir_output,
                 name="ADMM",
                 parameters={
                     "alpha": np.arange(0.01, 0.05, 0.01),
                     "rho": np.arange(0.1, 1.5, 0.5),
                     # "data_loss": ["linear", "arctan"],
                     # "data_loss_scale": [1., 1.2],
                 },
                 reconstruction_info={},
                 ):

        if not isinstance(solver, admm.ADMMLinearSolver):
            raise TypeError("solver must be of type 'ADMMLinearSolver'")

        super(self.__class__, self).__init__(
            solver=solver,
            parameters=parameters,
            observer=observer,
            dir_output=dir_output,
            name=name,
            reconstruction_info=reconstruction_info,
        )

    def _get_fileheader(self):

        # keys referring to the information to be printed in the file
        keys = ["alpha",
                "rho",
                "iterations",
                "minimizer",
                "iter_max",
                "x_scale",
                "data_loss",
                "data_loss_scale",
                "dimension",
                ]

        header = "## " + self._name
        for key in keys:
            # Only write information to header which stays constant in study
            if key not in self._parameters.keys():
                header += ", %s=%s" \
                    % (key, eval("str(self._solver.get_" + key + "())"))
        header += " (%s)" % (ph.get_time_stamp())
        header += "\n"
        return header



##
# \file run_deconvolution_test.py
#  \brief  Class to test deconvolution script
#
#  \author Michael Ebner (michael.ebner.14@ucl.ac.uk)
#  \date October 2017


import os
import unittest
import SimpleITK as sitk

import pysitk.python_helper as ph

from nsol.definitions import DIR_TMP, DIR_TEST


class RunDeconvolution(unittest.TestCase):

    def setUp(self):
        self.data_2d = os.path.join(DIR_TEST, "2D_Lena_256_blur_noise.png")
        self.result_2d = os.path.join(DIR_TMP, "2D_result.png")

        self.data_3d = os.path.join(DIR_TEST, "3D_SheppLoganPhantom_64.nii.gz")
        self.result_3d = os.path.join(DIR_TMP, "3D_result.nii.gz")

        self.iterations = 5

    def run_reconstruction_2d(self, reconstruction_type):
        cmd_args = []
        cmd_args.append("--observation %s" % self.data_2d)
        cmd_args.append("--result %s" % self.result_2d)
        cmd_args.append("--iterations %d" % self.iterations)
        cmd_args.append("--reconstruction-type %s" % reconstruction_type)
        cmd_args.append("--verbose 0")
        cmd = "nsol_run_deconvolution %s" % (" ").join(cmd_args)
        return ph.execute_command(cmd)

    def run_reconstruction_3d(self, reconstruction_type):
        cmd_args = []
        cmd_args.append("--observation %s" % self.data_3d)
        cmd_args.append("--result %s" % self.result_3d)
        cmd_args.append("--iterations %d" % self.iterations)
        cmd_args.append("--reconstruction-type %s" % reconstruction_type)
        cmd_args.append("--verbose 0")
        cmd = "nsol_run_deconvolution %s" % (" ").join(cmd_args)
        return ph.execute_command(cmd)

    def test_tkl2_deconvolution(self):

        # 2D
        self.assertEqual(self.run_reconstruction_2d("TK0L2"), 0)
        self.assertEqual(self.run_reconstruction_2d("TK1L2"), 0)

        # 3D
        self.assertEqual(self.run_reconstruction_3d("TK0L2"), 0)
        self.assertEqual(self.run_reconstruction_3d("TK1L2"), 0)

    def test_tvl2_deconvolution(self):

        # 2D
        self.assertEqual(self.run_reconstruction_2d("TVL2"), 0)

        # 3D
        self.assertEqual(self.run_reconstruction_3d("TVL2"), 0)

    def test_huberl2_deconvolution(self):

        # 2D
        self.assertEqual(self.run_reconstruction_2d("HuberL2"), 0)

        # 3D
        self.assertEqual(self.run_reconstruction_3d("HuberL2"), 0)



##
# \file noise.py
# \brief      Class to add different kinds of noise
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       July 2017
#

# Import libraries
import numpy as np


##
# Class to apply noise
# \date       2017-07-22 23:41:01+0100
#
class Noise(object):

    ##
    # Store data array where noise shall be applied
    # \date       2017-07-22 23:40:02+0100
    #
    # \param      self  The object
    # \param      data  Data as numpy array
    # \param      seed  The seed
    #
    def __init__(self, data, seed=None):
        np.random.seed(seed=seed)
        self._data = np.array(data)

    ##
    # Gets the noisy data.
    # \date       2017-07-22 23:41:17+0100
    #
    # \param      self  The object
    #
    # \return     The noisy data as numpy array
    #
    def get_noisy_data(self):
        return self._data

    ##
    # Adds additive Gaussian noise.
    # \date       2017-07-22 23:41:29+0100
    #
    # \param      self         The object
    # \param      noise_level  The noise level between 0 and 1
    # \param      mean         Mean/centre for Gaussian distribution
    # \param      sigma        Standard deviation for Gaussian
    #
    def add_gaussian_noise(self, noise_level=0.01, mean=0, sigma=1):
        self._data += noise_level * self._data.max() * \
            np.random.normal(size=self._data.shape, loc=mean, scale=sigma)

    ##
    # Adds additive Poisson noise.
    # \date       2017-07-22 23:41:29+0100
    #
    # \param      self         The object
    # \param      noise_level  The noise level between 0 and 1
    # \param      lmbda        Expectation of interval, float >= 0
    #
    def add_poisson_noise(self, noise_level=0.01, lmbda=1):
        self._data += noise_level * self._data.max() * \
            np.random.poisson(size=self._data.shape, lam=lmbda)

    ##
    # Adds an uniform noise.
    # \date       2017-07-22 23:42:11+0100
    #
    # \param      self         The object
    # \param      noise_level  The noise level between 0 and 1
    #
    def add_uniform_noise(self, noise_level=0.01):
        self._data += noise_level * self._data.max() * \
            np.random.rand(*self._data.shape)

    ##
    # Replace random pixels with zero (pepper) and one (salt)
    # \date       2017-07-22 22:51:41+0100
    #
    # \param      data          numpy array
    #
    # \return     noised numpy array
    #
    def add_salt_and_pepper_noise(self, salt_vs_pepper=0.5, amount=0.1):

        # Define value for salt and pepper
        val_salt = self._data.max()
        val_pepper = self._data.min()

        shape = self._data.shape
        self._data = self._data.flatten()

        # Pick size amount of random indices from data array
        size = int(amount * self._data.size)
        samples_all = np.random.choice(
            np.arange(0, self._data.size),
            size=size,
            replace=False)

        # Chose the first part to be of salt value
        N_white = int(salt_vs_pepper * samples_all.size)

        self._data[samples_all[0:N_white]] = val_salt
        self._data[samples_all[N_white:]] = val_pepper

        # Reshape to original shape
        self._data = self._data.reshape(*shape)



__title__ = "NSoL"
__author__ = "Michael Ebner"
__email__ = "michael.ebner.14@ucl.ac.uk"
__license__ = "BSD-3-Clause"
__summary__ = "The Numerical Solver Library (NSoL) is a research-focused " \
    "toolkit for denoising and deconvolution applications using ADMM and " \
    "Primal-Dual methods."
__uri__ = "https://github.com/gift-surg/NSoL"
__version__ = "0.1.6.2"



#!/usr/bin/python

##
# \file addNoise.py
# \brief      Tool to smooth data and add noise to it
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       Sept 2017
#

import os
import sys
import numpy as np
import SimpleITK as sitk

import pysitk.python_helper as ph
import pysitk.simple_itk_helper as sitkh

import nsol.input_argparser as InputArgparser
import nsol.data_reader as dr
import nsol.data_writer as dw
import nsol.linear_operators as LinearOperators
import nsol.noise as Noise

from nsol.definitions import ALLOWED_NOISE_TYPES


def main():

    input_parser = InputArgparser.InputArgparser(
        description="Tool to smooth data and add noise to it",
    )
    input_parser.add_filename(required=True)
    input_parser.add_result(required=True)
    input_parser.add_noise(default="gaussian")
    input_parser.add_noise_level(default=0.05)
    input_parser.add_blur(default=[1])
    args = input_parser.parse_args()
    input_parser.print_arguments(args)

    # --------------------------------Read Data--------------------------------
    data_reader = dr.DataReader(args.filename)
    data_reader.read_data()
    nda = data_reader.get_data()

    # --------------------------------Blur Data--------------------------------
    if args.blur[0] > 0:
        sigma = np.atleast_1d(args.blur)
        if sigma.ndim != nda.ndim:
            try:
                cov = np.diag(np.ones(nda.ndim)) * sigma**2
            except:
                raise IOError(
                    "Blur information must be either 1- or d-dimensional")

        linear_operators = eval(
            "LinearOperators.LinearOperators%dD()" % nda.ndim)
        A = linear_operators.get_gaussian_blurring_operators(cov)[0]
        nda = A(nda)

    # --------------------------------Add Noise--------------------------------
    noise = Noise.Noise(nda, seed=1)

    if args.noise == "gaussian":
        noise.add_gaussian_noise(noise_level=args.noise_level, sigma=1)

    elif args.noise == "poisson":
        noise.add_poisson_noise(noise_level=args.noise_level, lmbda=1)

    elif args.noise == "uniform":
        noise.add_uniform_noise(noise_level=args.noise_level)

    elif args.noise == "s&p":
        noise.add_salt_and_pepper_noise(salt_vs_pepper=0.5, amount=0.1)

    elif args.noise in [None, "none", "None"]:
        pass

    else:
        raise IOError("Noise type '%s' not known" % args.noise)

    nda = noise.get_noisy_data()

    # -------------------------------Write Data-------------------------------
    data_writer = dw.DataWriter(nda, args.result, data_reader.get_image_sitk())
    data_writer.write_data()

    return 0

if __name__ == '__main__':
    main()



##
# \file linear_operators.py
# \brief      Class to create linear operators used for blurring and
#             differentiation in both 2D and 3D
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       July 2017
#

# Import libraries
import numpy as np
import scipy.ndimage
from abc import ABCMeta, abstractmethod

import nsol.kernels as Kernels


class LinearOperators(object):
    __metaclass__ = ABCMeta

    ##
    # { constructor_description }
    # \date       2017-07-23 16:42:57+0100
    #
    # \param      self       The object
    # \param      dimension  Dimension of space, integer
    # \param      spacing    Spacing in each spatial direction in x- (y-, z-)
    #                        direction as numpy array or scalar (only 1D)
    #
    def __init__(self, dimension, spacing):

        self._dimension = dimension
        self._spacing = spacing

        if dimension == 1:
            self._kernels = Kernels.Kernels1D(spacing=spacing)
        elif dimension == 2:
            self._kernels = Kernels.Kernels2D(spacing=spacing)
        elif dimension == 3:
            self._kernels = Kernels.Kernels3D(spacing=spacing)

    def get_spacing(self):
        return self._spacing

    def get_dimension(self):
        return self._dimension

    ##
    # Gets the convolution and adjoint convolution operators given the provided
    # kernel
    # \date       2017-07-19 16:14:44+0100
    #
    # \param      self    The object
    # \param      kernel  Kernel specifying the linear operation as numpy array
    # \param      mode    Mode specifying the boundary conditions for the
    #                     convolution
    #
    # \return     The convolution and adjoint convolution operators.
    #
    def get_convolution_and_adjoint_convolution_operators(
            self, kernel, mode="wrap"):

        kernel_adj = kernel

        A = lambda x: scipy.ndimage.convolve(x, kernel, mode=mode)
        A_adj = lambda x: scipy.ndimage.convolve(x, kernel_adj, mode=mode)

        return A, A_adj

    ##
    # Gets the Gaussian blurring operator and its adjoint associated to a
    # covariance matrix.
    # \date       2017-07-19 16:16:18+0100
    #
    # \param      self       The object
    # \param      cov        Variance covariance matrix as numpy array
    # \param      alpha_cut  Cut-off distance in integer, i.e. 3 means cutting
    #                        off a 3 sigma in each direction
    #
    # \return     The gaussian blurring operators.
    #
    def get_gaussian_blurring_operators(self, cov, alpha_cut=3):

        kernel = self._kernels.get_gaussian(cov=cov, alpha_cut=alpha_cut)

        return self.get_convolution_and_adjoint_convolution_operators(kernel)

    ##
    # Gets the differential operator in x-direction and its adjoint
    # \date       2017-07-19 16:31:55+0100
    #
    # \param      self  The object
    # \param      mode  Mode specifying the boundary conditions for the
    #                   convolution
    #
    # \return     The x-differential operators.
    #
    def get_dx_operators(self, mode="constant"):

        kernel = self._kernels.get_dx_forward_difference()
        kernel_adj = -self._kernels.get_dx_backward_difference()

        D = lambda x: scipy.ndimage.convolve(x, kernel, mode=mode)
        D_adj = lambda x: scipy.ndimage.convolve(x, kernel_adj, mode=mode)

        return D, D_adj

    ##
    # Gets the gradient operator and its adjoint for both 2D and 3D.
    #
    # Operator \p grad applied on (m x n) numpy array returns an (dim*m x n)
    # numpy array, i.e. stacking the differentials on top of each other.
    # Operator \p grad_adj maps from (dim*m x n) to (m x n)
    # \date       2017-07-19 17:14:26+0100
    #
    # \param      self  The object
    # \param      mode  The mode
    #
    # \return     The gradient operators.
    #
    def get_gradient_operators(self, mode="constant"):

        Dx, Dx_adj = self.get_dx_operators(mode=mode)

        if self._dimension == 1:
            grad = Dx
            grad_adj = Dx_adj

        if self._dimension == 2:
            Dy, Dy_adj = self.get_dy_operators(mode=mode)

            grad = lambda x: np.concatenate((Dx(x), Dy(x)))
            grad_adj = lambda x: self._get_adjoint_gradient_operator(
                x, [Dx_adj, Dy_adj])

        elif self._dimension == 3:
            Dy, Dy_adj = self.get_dy_operators(mode=mode)
            Dz, Dz_adj = self.get_dz_operators(mode=mode)

            grad = lambda x: np.concatenate((Dx(x), Dy(x), Dz(x)))
            grad_adj = lambda x: self._get_adjoint_gradient_operator(
                x, [Dx_adj, Dy_adj, Dz_adj])

        return grad, grad_adj

    ##
    # Gets the adjoint gradient operator.
    #
    # Apply Dx_adj(x[0:m,...]) + Dy_adj(x[m:2m,...]) (+ Dz_adj(x[2m:,...]))
    # \date       2017-07-19 17:19:06+0100
    #
    # \param      self        The object
    # \param      x           numpy array of shape (dim*m x n x p)
    # \param      D_adj_list  The d adj list
    #
    # \return     The adjoint gradient operator.
    #
    def _get_adjoint_gradient_operator(self, x, D_adj_list):

        x_split = np.array_split(x, self._dimension)

        D_adj_x_list = [D_adj_list[i](x_split[i])
                        for i in range(0, self._dimension)]

        D_adj_x = D_adj_x_list[0]
        for i in range(1, self._dimension):
            D_adj_x += D_adj_x_list[i]

        return D_adj_x


class LinearOperators1D(LinearOperators):

    def __init__(self, spacing=1):
        super(self.__class__, self).__init__(dimension=1, spacing=spacing)


class LinearOperators2D(LinearOperators):

    def __init__(self, spacing=np.ones(2)):
        super(self.__class__, self).__init__(dimension=2, spacing=spacing)

    ##
    # Gets the differential operator in y-direction and its adjoint
    # \date       2017-07-19 16:31:55+0100
    #
    # \param      self  The object
    # \param      mode  Mode specifying the boundary conditions for the
    #                   convolution
    #
    # \return     The y-differential operators.
    #
    def get_dy_operators(self, mode="constant"):

        kernel = self._kernels.get_dy_forward_difference()
        kernel_adj = -self._kernels.get_dy_backward_difference()

        D = lambda x: scipy.ndimage.convolve(x, kernel, mode=mode)
        D_adj = lambda x: scipy.ndimage.convolve(x, kernel_adj, mode=mode)

        return D, D_adj


class LinearOperators3D(LinearOperators):

    def __init__(self, spacing=np.ones(3)):
        super(self.__class__, self).__init__(dimension=3, spacing=spacing)

    ##
    # Gets the differential operator in y-direction and its adjoint
    # \date       2017-07-19 16:31:55+0100
    #
    # \param      self  The object
    # \param      mode  Mode specifying the boundary conditions for the
    #                   convolution
    #
    # \return     The y-differential operators.
    #
    def get_dy_operators(self, mode="constant"):

        kernel = self._kernels.get_dy_forward_difference()
        kernel_adj = -self._kernels.get_dy_backward_difference()

        D = lambda x: scipy.ndimage.convolve(x, kernel, mode=mode)
        D_adj = lambda x: scipy.ndimage.convolve(x, kernel_adj, mode=mode)

        return D, D_adj

    ##
    # Gets the differential operator in z-direction and its adjoint for 3D
    # \date       2017-07-19 16:31:55+0100
    #
    # \param      self  The object
    # \param      mode  Mode specifying the boundary conditions for the
    #                   convolution
    #
    # \return     The z-differential operators.
    #
    def get_dz_operators(self, mode="constant"):

        kernel = self._kernels.get_dz_forward_difference()
        kernel_adj = -self._kernels.get_dz_backward_difference()

        D = lambda x: scipy.ndimage.convolve(x, kernel, mode=mode)
        D_adj = lambda x: scipy.ndimage.convolve(x, kernel_adj, mode=mode)

        return D, D_adj



##
# \file kernels_test.py
#  \brief  Class containing unit tests for module Kernels
#
#  \author Michael Ebner (michael.ebner.14@ucl.ac.uk)
#  \date July 2017

import numpy as np
import scipy
import unittest

import nsol.kernels as Kernels
import nsol.linear_operators as LinearOperators

from nsol.definitions import DIR_TEST


class KernelsTest(unittest.TestCase):

    def setUp(self):

        self.accuracy = 10
        self.scaling = 1

        # 1D
        self.shape_1D = (50,)

        self.cov_1D = 2

        self.x_1D = self.scaling * np.random.rand(*self.shape_1D)
        self.y_1D = self.scaling * np.random.rand(*self.shape_1D)

        self.linear_operators_1D = LinearOperators.LinearOperators1D()

        # 2D
        self.shape_2D = (50, 50)

        self.cov_2D = np.zeros((2, 2))
        self.cov_2D[0, 0] = 2
        self.cov_2D[1, 1] = 2

        self.x_2D = self.scaling * np.random.rand(*self.shape_2D)
        self.y_2D = self.scaling * np.random.rand(*self.shape_2D)

        self.linear_operators_2D = LinearOperators.LinearOperators2D()

        # 3D
        self.shape_3D = (50, 50, 10)
        self.cov_3D = np.zeros((3, 3))
        self.cov_3D[0, 0] = 2
        self.cov_3D[1, 1] = 2
        self.cov_3D[2, 2] = 2

        self.x_3D = self.scaling * np.random.rand(*self.shape_3D)
        self.y_3D = self.scaling * np.random.rand(*self.shape_3D)

        self.linear_operators_3D = LinearOperators.LinearOperators3D()

    def test_kernel_spacings(self):
        # ---------------------------------1D----------------------------------
        spacing = 2
        kernel = Kernels.Kernels1D(spacing=spacing)

        # dx
        k = kernel.get_dx_forward_difference()
        abs_diff = np.linalg.norm(
            np.abs(k[k.nonzero()]) - 1./spacing)
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        k = kernel.get_dx_backward_difference()
        abs_diff = np.linalg.norm(
            np.abs(k[k.nonzero()]) - 1./spacing)
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        # ---------------------------------2D----------------------------------
        spacing = np.array([2, 4])
        kernel = Kernels.Kernels2D(spacing=spacing)

        # dx
        k = kernel.get_dx_forward_difference()
        abs_diff = np.linalg.norm(
            np.abs(k[k.nonzero()]) - 1./spacing[0])
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        k = kernel.get_dx_backward_difference()
        abs_diff = np.linalg.norm(
            np.abs(k[k.nonzero()]) - 1./spacing[0])
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        # dy
        k = kernel.get_dy_forward_difference()
        abs_diff = np.linalg.norm(
            np.abs(k[k.nonzero()]) - 1./spacing[1])
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        k = kernel.get_dy_backward_difference()
        abs_diff = np.linalg.norm(
            np.abs(k[k.nonzero()]) - 1./spacing[1])
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        # ---------------------------------3D----------------------------------
        spacing = np.array([2, 4, 6])
        kernel = Kernels.Kernels3D(spacing=spacing)

        # dx
        k = kernel.get_dx_forward_difference()
        abs_diff = np.linalg.norm(
            np.abs(k[k.nonzero()]) - 1./spacing[0])
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        k = kernel.get_dx_backward_difference()
        abs_diff = np.linalg.norm(
            np.abs(k[k.nonzero()]) - 1./spacing[0])
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        # dy
        k = kernel.get_dy_forward_difference()
        abs_diff = np.linalg.norm(
            np.abs(k[k.nonzero()]) - 1./spacing[1])
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        k = kernel.get_dy_backward_difference()
        abs_diff = np.linalg.norm(
            np.abs(k[k.nonzero()]) - 1./spacing[1])
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        # dz
        k = kernel.get_dz_forward_difference()
        abs_diff = np.linalg.norm(
            np.abs(k[k.nonzero()]) - 1./spacing[2])
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        k = kernel.get_dz_backward_difference()
        abs_diff = np.linalg.norm(
            np.abs(k[k.nonzero()]) - 1./spacing[2])
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

    def test_gaussian_blurring(self):

        # ---------------------------------1D----------------------------------
        A, A_adj = self.linear_operators_1D.get_gaussian_blurring_operators(
            self.cov_1D)

        A_x = A(self.x_1D)
        A_adj_y = A_adj(self.y_1D)

        # Check |(Ax,y) - (x,A'y)| = 0
        abs_diff = np.abs(np.sum(A_x*self.y_1D) - np.sum(self.x_1D*A_adj_y))
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        # ---------------------------------2D----------------------------------
        A, A_adj = self.linear_operators_2D.get_gaussian_blurring_operators(
            self.cov_2D)

        A_x = A(self.x_2D)
        A_adj_y = A_adj(self.y_2D)

        # Check |(Ax,y) - (x,A'y)| = 0
        abs_diff = np.abs(np.sum(A_x*self.y_2D) - np.sum(self.x_2D*A_adj_y))
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        # ---------------------------------3D----------------------------------
        A, A_adj = self.linear_operators_3D.get_gaussian_blurring_operators(
            self.cov_3D)
        A_x = A(self.x_3D)
        A_adj_y = A_adj(self.y_3D)

        # Check |(Ax,y) - (x,A'y)| = 0
        abs_diff = np.abs(np.sum(A_x*self.y_3D) - np.sum(self.x_3D*A_adj_y))
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

    def test_differential_operations(self):

        # ---------------------------------1D----------------------------------

        # dx
        D, D_adj = self.linear_operators_1D.get_dx_operators()
        D_x = D(self.x_1D)
        D_adj_y = D_adj(self.y_1D)
        abs_diff = np.abs(np.sum(D_x*self.y_1D) - np.sum(self.x_1D*D_adj_y))
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        # ---------------------------------2D----------------------------------

        # dx
        D, D_adj = self.linear_operators_2D.get_dx_operators()
        D_x = D(self.x_2D)
        D_adj_y = D_adj(self.y_2D)
        abs_diff = np.abs(np.sum(D_x*self.y_2D) - np.sum(self.x_2D*D_adj_y))
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        # dy
        D, D_adj = self.linear_operators_2D.get_dy_operators()
        D_x = D(self.x_2D)
        D_adj_y = D_adj(self.y_2D)
        abs_diff = np.abs(np.sum(D_x*self.y_2D) - np.sum(self.x_2D*D_adj_y))
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        # ---------------------------------3D----------------------------------

        # dx
        D, D_adj = self.linear_operators_3D.get_dx_operators()
        D_x = D(self.x_3D)
        D_adj_y = D_adj(self.y_3D)
        abs_diff = np.abs(np.sum(D_x*self.y_3D) - np.sum(self.x_3D*D_adj_y))
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        # dy
        D, D_adj = self.linear_operators_3D.get_dy_operators()
        D_x = D(self.x_3D)
        D_adj_y = D_adj(self.y_3D)
        abs_diff = np.abs(np.sum(D_x*self.y_3D) - np.sum(self.x_3D*D_adj_y))
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        # dz
        D, D_adj = self.linear_operators_3D.get_dz_operators()
        D_x = D(self.x_3D)
        D_adj_y = D_adj(self.y_3D)
        abs_diff = np.abs(np.sum(D_x*self.y_3D) - np.sum(self.x_3D*D_adj_y))
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

    def test_get_gradient_operators(self):

        # ---------------------------------1D----------------------------------
        dx, dx_adj = self.linear_operators_1D.get_dx_operators()
        grad, grad_adj = self.linear_operators_1D.get_gradient_operators()

        # grad
        x = self.x_1D
        y0 = dx(x)
        res = y0
        res_ = grad(x)
        abs_diff = np.abs(np.sum(res_-res))
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        # grad_adj
        y = self.scaling * np.random.rand(*self.shape_1D)

        y0 = y
        res = dx_adj(y0)
        res_ = grad_adj(y)
        abs_diff = np.abs(np.sum(res_-res))
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        # ---------------------------------2D----------------------------------
        dx, dx_adj = self.linear_operators_2D.get_dx_operators()
        dy, dy_adj = self.linear_operators_2D.get_dy_operators()
        grad, grad_adj = self.linear_operators_2D.get_gradient_operators()

        # grad
        x = self.x_2D
        y0 = dx(x)
        y1 = dy(x)
        res = np.concatenate((y0, y1))
        res_ = grad(x)
        abs_diff = np.abs(np.sum(res_-res))
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        # grad_adj
        y = self.scaling * np.concatenate((
            np.random.rand(*self.shape_2D),
            np.random.rand(*self.shape_2D)))

        m = self.shape_2D[0]
        y0 = y[:m, :]
        y1 = y[m:2*m, :]
        res = dx_adj(y0) + dy_adj(y1)
        res_ = grad_adj(y)
        abs_diff = np.abs(np.sum(res_-res))
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        # ---------------------------------3D----------------------------------
        dx, dx_adj = self.linear_operators_3D.get_dx_operators()
        dy, dy_adj = self.linear_operators_3D.get_dy_operators()
        dz, dz_adj = self.linear_operators_3D.get_dz_operators()
        grad, grad_adj = self.linear_operators_3D.get_gradient_operators()

        # grad
        x = self.x_3D
        y0 = dx(x)
        y1 = dy(x)
        y2 = dz(x)
        res = np.concatenate((y0, y1, y2))
        res_ = grad(x)
        abs_diff = np.abs(np.sum(res_-res))
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        # grad_adj
        y = self.scaling * np.concatenate((
            0*np.random.rand(*self.shape_3D),
            0*np.random.rand(*self.shape_3D),
            np.random.rand(*self.shape_3D)))

        m = self.shape_3D[0]
        y0 = y[:m, ...]
        y1 = y[m:2*m, ...]
        y2 = y[2*m:, ...]
        res = dx_adj(y0) + dy_adj(y1) + dz_adj(y2)
        res_ = grad_adj(y)
        abs_diff = np.abs(np.sum(res_-res))
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

    def test_gradient_operators(self):

        # ---------------------------------1D-------------------------------

        grad, grad_adj = self.linear_operators_1D.get_gradient_operators()
        x = self.x_1D
        y = self.scaling * np.random.rand(*self.shape_1D)

        abs_diff = np.abs(np.sum(grad(x)*y) - np.sum(x*grad_adj(y)))
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        # ---------------------------------2D-------------------------------

        grad, grad_adj = self.linear_operators_2D.get_gradient_operators()
        x = self.x_2D
        y = self.scaling * np.concatenate((
            np.random.rand(*self.shape_2D),
            np.random.rand(*self.shape_2D)))

        abs_diff = np.abs(np.sum(grad(x)*y) - np.sum(x*grad_adj(y)))
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)

        # ---------------------------------3D-------------------------------

        grad, grad_adj = self.linear_operators_3D.get_gradient_operators()
        x = self.x_3D
        y = self.scaling * np.concatenate((
            np.random.rand(*self.shape_3D),
            np.random.rand(*self.shape_3D),
            np.random.rand(*self.shape_3D)))

        abs_diff = np.abs(np.sum(grad(x)*y) - np.sum(x*grad_adj(y)))
        self.assertEqual(np.round(abs_diff, decimals=self.accuracy), 0)



##
# \file proximalOperators.py
# \brief      Collection of proximal operators for primal and dual variables
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       July 2017
#

import numpy as np

import nsol.tikhonov_linear_solver as tk


class ProximalOperators(object):

    # #########################################################################
    # ------------------------------Primal Variables--------------------------

    ##
    # Proximal operator for g(x) = 1/2 ||Ax-b||^2
    #
    # Approximation provided by solving the minimization problem via Tikhonov
    # solver
    # \date       2017-07-23 01:32:15+0100
    #
    # \param      x         1D numpy array
    # \param      tau       step size; scalar >= 0
    # \param      A         Function handle lambda x: A(x) with x being 1D
    #                       numpy array
    # \param      A_adj     Function handle for adjoint operator lambda x:
    #                       A^*(x)
    # \param      b         1D numpy array
    # \param      x0        initial value for minimization problem as 1D numpy
    #                       array
    # \param      iter_max  The iterator maximum
    # \param      verbose   Verbose output; bool
    # \param      x_scale   Characteristic scale of each variable. Setting
    #                       x_scale is equivalent to reformulating the problem
    #                       in scaled variables ``xs = x / x_scale``
    #
    # \return     prox_g(x, tau) as numpy array
    #
    @staticmethod
    def prox_linear_least_squares(x,
                                  tau,
                                  A,
                                  A_adj,
                                  b,
                                  x0,
                                  iter_max=10,
                                  verbose=0,
                                  data_loss="linear",
                                  data_loss_scale=1,
                                  minimizer="lsmr",
                                  x_scale=1,
                                  bounds=(0, np.inf),
                                  ):

        # Identity
        I = lambda x: x.flatten()

        tikhonov = tk.TikhonovLinearSolver(
            A=A, A_adj=A_adj,
            B=I, B_adj=I,
            x0=x0/float(x_scale),
            b=b/float(x_scale),
            b_reg=x,
            alpha=1./tau,
            iter_max=iter_max,
            verbose=verbose,
            x_scale=x_scale,
            data_loss=data_loss,
            data_loss_scale=data_loss_scale,
            minimizer=minimizer,
            bounds=bounds,
        )
        tikhonov.run()
        return tikhonov.get_x()

    ##
    # Proximal operator for ell1 denoising, i.e. g(x) = ||x - x0||_1.
    #
    # See Chambolle2011, p.135
    # \date       2017-07-21 01:52:01+0100
    #
    # \param      x        numpy array
    # \param      tau      step size; scalar >= 0
    # \param      x0       observed; numpy array
    # \param      x_scale  Characteristic scale of each variable. Setting
    #                      x_scale is equivalent to reformulating the problem
    #                      in scaled variables ``xs = x / x_scale``
    #
    # \return     prox_{\tau g}(x) as numpy array
    #
    @staticmethod
    def prox_ell1_denoising(x, tau, x0, x_scale=1.):
        x0 = x0 / float(x_scale)
        return x0 + np.maximum(np.abs(x-x0) - tau, 0) * np.sign(x-x0)

    ##
    # Proximal operator for least squares denoising, i.e. g(x) = 1/2 ||x -
    # x0||_2^2
    #
    # With g(x) = 1/2 * ||x - x0||_2^2 it holds that prox_{\tau g}(x) = (x +
    # \tau*x0) / (1 + \tau*x0). See Chambolle2011, p.133
    # \date       2017-07-21 01:52:01+0100
    #
    # \param      x        numpy array
    # \param      tau      step size; scalar >= 0
    # \param      x0       observed; numpy array
    # \param      x_scale  Characteristic scale of each variable. Setting
    #                      x_scale is equivalent to reformulating the problem
    #                      in scaled variables ``xs = x / x_scale``
    #
    # \return     prox_{\tau g}(x) as numpy array
    #
    @staticmethod
    def prox_ell2_denoising(x, tau, x0, x_scale=1.):
        x0 = x0 / float(x_scale)
        return (x + tau * x0) / (1. + tau)

    # #########################################################################
    # -------------------------------Dual Variables---------------------------

    ##
    # Proximal operator for conjugate of TV functional.
    #
    # With g(x) = TV(x) = || ||\nabla x||_2 ||_1 the conjugate operator reads
    # g'(p) = \delta_P(p)
    # See Chambolle2011, p.133
    # \date       2017-07-23 17:58:04+0100
    #
    # \param      x      numpy array
    # \param      sigma  step size dual variable; scalar >= 0
    #
    # \return     prox_{\sigma g'}(x) as numpy array
    #
    @staticmethod
    def prox_tv_conj(x, sigma):
        return x / np.maximum(1, np.abs(x))

    ##
    # Proximal operator for conjugate of Huber functional.
    #
    # With g(x) = | ||\nabla x||_2 |_\gamma the conjugate operator reads
    # g'(p) = \delta_P(p) + gamma/2 ||p||_2^2
    # See Chambolle2011. p.136
    # \date       2017-07-23 18:03:56+0100
    #
    #
    # \param      x      numpy array
    # \param      sigma  step size dual variable; scalar >= 0
    #
    # \return     prox_{\sigma g'}(x) as numpy array
    #
    @staticmethod
    def prox_huber_conj(x, sigma, gamma=0.05):
        x /= (1. + sigma * gamma)
        return x / np.maximum(1, np.abs(x))



##
# \file admm_linear_solver.py
# \brief      Class to define a numerical solver for solving the linear
#             least-squares problems with TV regularization via the Alternating
#             Direction Method of Multipliers (ADMM) method
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       July 2017
#

import os
import sys
import numpy as np

import pysitk.python_helper as ph

from nsol.linear_solver import LinearSolver
import nsol.tikhonov_linear_solver as tk
from nsol.prior_measures import PriorMeasures as prior_meas


##
# Class to estimate the unique minimizer to the convex minimization problem
# max_x [1/2 ||rho( Ax-b )||^2 + alpha g(Bx-b_reg)] with g(z) = TV(z) using
# ADMM.
# \date       2017-07-21 00:23:34+0100
#
class ADMMLinearSolver(LinearSolver):

    ##
    # Store relevant information for solving a linear least-squares problem
    # with TV regularization via ADMM
    # \date       2017-07-21 00:24:24+0100
    #
    # \param      self             The object
    # \param      A                Function associated with linear operator A:
    #                              X->Y; x->A(x) with x being a 1D numpy array
    # \param      A_adj            Function associated with adjoint linear
    #                              operator A^*: Y->X; y->A^*(y)
    # \param      b                Right hand-side of linear system Ax = b as
    #                              1D numpy array
    # \param      B                Function associated with linear operator B:
    #                              x->B(x) with x and B(x) 1D numpy arrays
    # \param      B_adj            Function associated with adjoint linear
    #                              operator B^*
    # \param      x0               Initial value as 1D numpy array
    # \param      dimension        Dimension of space as integer indicating
    #                              either 1D, 2D or 3D problems
    # \param      b_reg            Right hand-side of linear system associated
    #                              to the regularizer, i.e. Bx = b_reg.
    # \param      alpha            Regularization parameter; scalar > 0
    # \param      iter_max         Number of maximum iterations for used
    #                              minimizer, integer value
    # \param      minimizer        String defining the used optimizer, i.e.
    #                              "lsmr", "least_squares" or any solver as
    #                              provided by scipy.optimize.minimize
    # \param      data_loss        Data loss function rho specified as string,
    #                              e.g. "linear", "soft_l1", "huber", "cauchy",
    #                              "arctan".
    # \param      data_loss_scale  Value of soft margin between inlier and
    #                              outlier residuals, default is 1.0. The loss
    #                              function is evaluated as rho_(f2) = C**2 *
    #                              rho(f2 / C**2), where C is data_loss_scale.
    #                              This parameter has no effect with
    #                              data_loss='linear', but for other loss
    #                              values it is of crucial importance.
    # \param      rho              regularization parameter of augmented
    #                              Lagrangian term; scalar > 0
    # \param      iterations       Number of ADMM iterations, integer value
    # \param      x_scale          Characteristic scale of each variable.
    #                              Setting x_scale is equivalent to
    #                              reformulating the problem in scaled
    #                              variables ``xs = x / x_scale``
    # \param      verbose          Verbose output, bool
    #
    def __init__(self,
                 A, A_adj,
                 b,
                 B, B_adj,
                 x0,
                 dimension,
                 b_reg=0,
                 alpha=0.01,
                 iter_max=10,
                 minimizer="lsmr",
                 data_loss="linear",
                 data_loss_scale=1,
                 rho=0.5,
                 iterations=10,
                 x_scale=1,
                 verbose=0):

        super(self.__class__, self).__init__(
            A=A, A_adj=A_adj, b=b, x0=x0, alpha=alpha, iter_max=iter_max,
            minimizer=minimizer, data_loss=data_loss,
            data_loss_scale=data_loss_scale, x_scale=x_scale, verbose=verbose)

        self._B = B
        self._B_adj = B_adj
        self._b_reg = b_reg / self._x_scale
        self._dimension = dimension
        self._rho = float(rho)
        self._iterations = iterations

    ##
    # Sets the regularization parameter for augmented Lagrangian.
    # \date       2017-08-05 04:08:03+0100
    #
    # \param      self  The object
    # \param      rho   regularization  parameter of augmented Lagrangian term;
    #                   scalar > 0
    #
    def set_rho(self, rho):
        self._rho = rho

    ##
    # Gets the regularization parameter for augmented Lagrangian.
    # \date       2017-08-05 04:08:42+0100
    #
    # \param      self  The object
    #
    # \return     scalar > 0
    #
    def get_rho(self):
        return self._rho

    ##
    # Gets the dimension of space.
    # \date       2017-08-05 18:42:33+0100
    #
    # \param      self  The object
    #
    # \return     The dimension of space as int.
    #
    def get_dimension(self):
        return self._dimension

    ##
    # Sets the number of ADMM iterations.
    # \date       2017-08-05 18:59:12+0100
    #
    # \param      self        The object
    # \param      iterations  Number of ADMM iterations, integer value
    #
    def set_iterations(self, iterations):
        self._iterations = iterations

    ##
    # Gets the number of ADMM iterations.
    # \date       2017-08-05 18:59:39+0100
    #
    # \param      self  The object
    #
    # \return     Number of ADMM iterations, integer value.
    #
    def get_iterations(self):
        return self._iterations

    ##
    # Execute solver
    # \date       2017-08-05 19:00:53+0100
    #
    # \param      self  The object
    #
    def _run(self):

        # Monitor output
        if self._observer is not None:
            self._observer.add_x(self.get_x())

        v = self._B(self._x0) - self._b_reg
        w = np.zeros_like(v)

        for i in range(0, self._iterations):

            if self._verbose:
                ph.print_title("ADMM iteration %d/%d" %
                               (i+1, self._iterations))
            else:
                ph.print_info("ADMM iteration %d/%d" %
                              (i+1, self._iterations))

            self._x, v, w = self._perform_ADMM_iteration(self._x, v, w)

            # Monitor output
            if self._observer is not None:
                self._observer.add_x(self.get_x())

            # shape = (256, 256)

            # recon = np.array(self._x.reshape(*shape))
            # ph.show_arrays(recon, title="ADMM_iter%d"%(i+1), fig_number=1)

            # v_split = [h.reshape(*shape) for h in self._get_split(v, self._dimension)]
            # ph.show_arrays(v_split, title="v", cmap="jet", fig_number=2)

            # w_split = [h.reshape(*shape) for h in self._get_split(w, self._dimension)]
            # ph.show_arrays(w_split, title="w", cmap="jet", fig_number=3)

            # ph.pause()

    def _perform_ADMM_iteration(self, x, v, w):

        # 1) Update primal variable
        x = self._solve_tikhonov_least_squares(x, v, w)

        # Compute derivatives for steps 2 and 3
        Bx_plus_w_minus_b_reg = self._B(x) + w - self._b_reg

        # 2) Update auxiliary variable
        v = self._prox_g(Bx_plus_w_minus_b_reg,
                         tau=self._alpha / self._rho,
                         dimension=self._dimension)

        # 3) Update scaled dual variable
        w = Bx_plus_w_minus_b_reg - v

        return x, v, w

    def _solve_tikhonov_least_squares(self, x, v, w):

        b_reg = v - w + self._b_reg

        tikhonov = tk.TikhonovLinearSolver(
            A=self._A, A_adj=self._A_adj,
            B=self._B, B_adj=self._B_adj,
            b=self._b, b_reg=b_reg,
            alpha=self._rho,
            x0=x,
            x_scale=1,
            iter_max=self._iter_max,
            data_loss=self._data_loss,
            minimizer=self._minimizer,
            verbose=self._verbose)
        tikhonov.run()

        return tikhonov.get_x()

    def _prox_g(self, t, tau, dimension):
        t_split = self._get_split(t, dimension)
        t_norm = np.sqrt(self._get_squared_sum_of_split(t_split))

        ind = t_norm > tau

        v = np.zeros_like(t)
        m = t_split[0].shape[0]
        for i in range(0, dimension):
            v_tmp = v[i*m:(i+1)*m, ...]
            v_tmp[ind] = self._get_soft_threshold(tau, t_norm[ind]) * \
                t_split[i][ind] / t_norm[ind]
            v[i*m:(i+1)*m, ...] = v_tmp

        return v

    ##
    # Gets the split of (n, ...) numpy into d (n/d, ...) numpy arrays with d
    # being the dimension of space
    #
    # This recovers the respective x (y, z) variables from a single numpy array
    # \date       2017-07-21 00:38:54+0100
    #
    # \param      self       The object
    # \param      x          Numpy array
    # \param      dimension  Numpy array as integer
    #
    # \return     List of d numpy arrays corresponding to the split
    #
    def _get_split(self, x, dimension):
        x_split = np.array_split(x, dimension)
        return x_split

    ##
    # Given a split, return the squared sum
    # \date       2017-07-21 00:41:49+0100
    #
    # \param      self     The object
    # \param      x_split  List of split obtained via 'get_split'
    #
    # \return     The squared sum of the split.
    #
    def _get_squared_sum_of_split(self, x_split):

        # x_i ** 2
        tmp = x_split[0] ** 2

        # Compute sum_k x_k^2
        for i in range(1, len(x_split)):
            tmp += x_split[i]**2

        return tmp

    ##
    # Gets the soft threshold.
    #
    # The soft threshold is defined as
    # \f[ S_\ell(t) =  \max(|t|-\ell,0)\,\text{sgn}(t) = \begin{cases}
    # t-\ell,& \text{if } t>\ell \\ 0,& \text{if } |t|\le\ell \\ t+\ell,&
    # \text{if } t<\ell \end{cases}
    # \f]
    # \date       2017-07-21 00:37:13+0100
    #
    # \param      self  The object
    # \param      ell   threshold as scalar > 0
    # \param      t     array containing the values to be thresholded
    #
    # \return     The soft threshold.
    #
    def _get_soft_threshold(self, ell, t):
        return np.maximum(np.abs(t) - ell, 0) * np.sign(t)

    def _get_cost_regularization_term(self, x):
        return prior_meas.total_variation(x, self._B, self._dimension)



##
# \file loss_functions.py
# \brief      Collection of loss functions and mapping from residual to cost
#
# The overall cost function is defined as
# \f[ C:\,\mathbb{R}^n \rightarrow \mathbb{R}_{\ge 0},\,\vec{x} \mapsto
# C(\vec{x}) := \frac{1}{2} \sum_{i=0}^{m-1} \rho( f_i(\vec{x})^2 )
# \f] with the loss function
# \f$\rho:\,\mathbb{R}\rightarrow\mathbb{R}_{\ge0}\f$ and the residual
# \f$ \vec{f}:\, \mathbb{R}^n \rightarrow
# \mathbb{R}^m,\,\vec{x}\mapsto\vec{f}(\vec{x}) = \b
# ig(f_0(\vec{x}),\,f_1(\vec{x}),\dots, f_{m-1}(\vec{x})\big)
# \f$
#
# \see        https://docs.scipy.org/doc/scipy-0.19.0/reference/generated/scipy.optimize.least_squares.html
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       July 2017
#


import numpy as np


class LossFunctions(object):

    # get_loss = {}
    # get_gradient_loss = {}

    ##
    # Gets the cost
    # \f$C(\vec{x})\f$ from the residual
    # \f$\vec{f}(\vec{x})\f$.
    # \date       2017-07-14 11:42:18+0100
    #
    # \param      f        residual f, m-dimensional numpy array
    # \param      loss     Choice for loss function rho
    # \param      f_scale  The f scale
    #
    # \return     The cost from residual as scalar value >= 0
    #
    @staticmethod
    def get_ell2_cost_from_residual(f, loss="linear", f_scale=1.):
        cost = 0.5 * \
            np.sum(LossFunctions.get_loss[loss](f2=f**2, f_scale=f_scale))
        return cost

    ##
    # Gets the gradient
    # \f$
    # \nabla C\f$ of the cost function given the residual
    # \f$\vec{f}(\vec{x})\f$ and its Jacobian
    # \f$\frac{d\vec{f}}{d\vec{x}}(\vec{x})\f$.
    # \date       2017-07-14 11:30:11+0100
    #
    # \param      f        residual f, m-dimensional numpy array
    # \param      jac_f    Jacobian of residual f, (m x n)-dimensional numpy
    #                      array
    # \param      loss     Choice for loss function rho
    # \param      f_scale  The f scale
    #
    # \return     The gradient of the cost from residual as n-dimensional numpy
    #             array
    #
    @staticmethod
    def get_gradient_ell2_cost_from_residual(f, jac_f, loss="linear",
                                             f_scale=1.):
        grad = np.sum((LossFunctions.get_gradient_loss[loss](
            f2=f**2,
            f_scale=f_scale) * f)[:, np.newaxis] * jac_f,
            axis=0)
        return grad

    ##
    # Return linear loss, i.e. rho(f2) = f2
    # \date       2017-07-25 19:34:23+0100
    #
    # \param      f2    scalar or numpy array; meant to be squared residual
    #
    # \return     rho(f2) as scalar or numpy array
    #
    @staticmethod
    def linear(f2, f_scale=1.):
        return f2

    ##
    # Return gradient/derivative of linear loss, i.e. rho'(f2) = 1
    # \date       2017-07-25 19:36:29+0100
    #
    # \param      f2    scalar or numpy array; meant to be squared residual
    #
    # \return     rho'(f2) as scalar or numpy array
    #
    @staticmethod
    def gradient_linear(f2, f_scale=1.):
        return np.ones_like(f2).astype(np.float64)

    ##
    # Return soft ell1 approximation, i.e. rho(f2) = 2 (sqrt(1 + f2) - 1)
    # \date       2017-07-25 19:37:52+0100
    #
    # \param      f2       scalar or numpy array; meant to be squared residual
    # \param      f_scale  Value of soft margin between inlier and outlier
    #                      residuals, default is 1.0. The loss function is
    #                      evaluated as rho_(f2) = C**2 * rho(f2 / C**2), where
    #                      C is f_scale. It is of crucial importance.
    #
    # \return     rho(f2) as scalar or numpy array
    #
    @staticmethod
    def soft_l1(f2, f_scale=1.):
        f_scale2 = float(f_scale * f_scale)
        f2 = f2 / f_scale2
        return 2. * (np.sqrt(1. + f2) - 1.) * f_scale2

    ##
    # Return gradient/derivative of soft ell1 loss, i.e. rho'(f2) = 1 /
    # sqrt(1+f2)
    # \date       2017-07-25 19:36:29+0100
    #
    # \param      f2       scalar or numpy array; meant to be squared residual
    # \param      f_scale  Value of soft margin between inlier and outlier
    #                      residuals, default is 1.0. The loss function is
    #                      evaluated as rho_'(f2) = rho'(f2 / C**2), where C is
    #                      f_scale. It is of crucial importance.
    #
    # \return     rho'(f2) as scalar or numpy array
    #
    @staticmethod
    def gradient_soft_l1(f2, f_scale=1.):
        f_scale2 = float(f_scale * f_scale)
        f2 = f2 / f_scale2
        return 1. / np.sqrt(1. + f2)

    ##
    # Return huber loss
    # \date       2017-07-25 19:37:52+0100
    #
    # \param      f2       scalar or numpy array; meant to be squared residual
    # \param      gamma    scalar > 0;
    # \param      f_scale  Value of soft margin between inlier and outlier
    #                      residuals, default is 1.0. The loss function is
    #                      evaluated as rho_(f2) = C**2 * rho(f2 / C**2), where
    #                      C is f_scale. It is of crucial importance.
    #
    # \return     rho(f2) as scalar or numpy array
    #
    @staticmethod
    def huber(f2, gamma=1.345, f_scale=1.):
        gamma = float(gamma)
        gamma2 = gamma * gamma
        f_scale2 = float(f_scale * f_scale)
        f2 = f2 / f_scale2
        return np.where(f2 < gamma2,
                        f2,
                        2.*gamma*np.sqrt(f2)-gamma2) * f_scale2

    ##
    # Return gradient/derivative of linear loss, i.e. rho'(f2)
    # \date       2017-07-25 19:36:29+0100
    #
    # \param      f2    scalar or numpy array; meant to be squared residual
    # \param      f_scale  Value of soft margin between inlier and outlier
    #                      residuals, default is 1.0. The loss function is
    #                      evaluated as rho_'(f2) = rho'(f2 / C**2), where C is
    #                      f_scale. It is of crucial importance.
    #
    # \return     rho'(f2) as scalar or numpy array
    #
    @staticmethod
    def gradient_huber(f2, gamma=1.345, f_scale=1.):
        gamma = float(gamma)
        gamma2 = gamma * gamma
        f_scale2 = float(f_scale * f_scale)
        f2 = f2 / f_scale2
        return np.where(f2 < gamma2, 1., gamma/np.sqrt(f2))

    ##
    # Return cauchy loss, i.e. rho(f2) = ln(1 + f2)
    # \date       2017-07-25 19:37:52+0100
    #
    # \param      f2       scalar or numpy array; meant to be squared residual
    # \param      f_scale  Value of soft margin between inlier and outlier
    #                      residuals, default is 1.0. The loss function is
    #                      evaluated as rho_(f2) = C**2 * rho(f2 / C**2), where
    #                      C is f_scale. It is of crucial importance.
    #
    # \return     rho(f2) as scalar or numpy array
    #
    @staticmethod
    def cauchy(f2, f_scale=1.):
        f_scale2 = float(f_scale * f_scale)
        f2 = f2 / f_scale2
        return np.log1p(f2) * f_scale2

    ##
    # Return gradient/derivative of cauchy loss, i.e. rho'(f2) = 1 / (1 + f2)
    # \date       2017-07-25 19:36:29+0100
    #
    # \param      f2       scalar or numpy array; meant to be squared residual
    # \param      f_scale  Value of soft margin between inlier and outlier
    #                      residuals, default is 1.0. The loss function is
    #                      evaluated as rho_'(f2) = rho'(f2 / C**2), where C is
    #                      f_scale. It is of crucial importance.
    #
    # \return     rho'(f2) as scalar or numpy array
    #
    @staticmethod
    def gradient_cauchy(f2, f_scale=1.):
        f_scale2 = float(f_scale * f_scale)
        f2 = f2 / f_scale2
        return 1. / (1. + f2)

    ##
    # Return arctan loss, i.e. rho(f2) = arctan(f2)
    # \date       2017-07-25 19:37:52+0100
    #
    # \param      f2       scalar or numpy array; meant to be squared residual
    # \param      f_scale  Value of soft margin between inlier and outlier
    #                      residuals, default is 1.0. The loss function is
    #                      evaluated as rho_(f2) = C**2 * rho(f2 / C**2), where
    #                      C is f_scale. It is of crucial importance.
    #
    # \return     rho(f2) as scalar or numpy array
    #
    @staticmethod
    def arctan(f2, f_scale=1.):
        f_scale2 = float(f_scale * f_scale)
        f2 = f2 / f_scale2
        return np.arctan(f2) * f_scale2

    ##
    # Return gradient/derivative of arctan loss, i.e. rho'(f2) = 1 / (1 + f2^2)
    # \date       2017-07-25 19:36:29+0100
    #
    # \param      f2       scalar or numpy array; meant to be squared residual
    # \param      f_scale  Value of soft margin between inlier and outlier
    #                      residuals, default is 1.0. The loss function is
    #                      evaluated as rho_'(f2) = rho'(f2 / C**2), where C is
    #                      f_scale. It is of crucial importance.
    #
    # \return     rho'(f2) as scalar or numpy array
    #
    @staticmethod
    def gradient_arctan(f2, f_scale=1.):
        f_scale2 = float(f_scale * f_scale)
        f2 = f2 / f_scale2
        return 1. / (1. + f2**2)

    # Dictionary for all loss functions
    get_loss = {
        "linear": linear.__func__,
        "soft_l1": soft_l1.__func__,
        "huber": huber.__func__,
        "cauchy": cauchy.__func__,
        "arctan": arctan.__func__,
    }

    # Dictionary for all gradient loss functions
    get_gradient_loss = {
        "linear": gradient_linear.__func__,
        "soft_l1": gradient_soft_l1.__func__,
        "huber": gradient_huber.__func__,
        "cauchy": gradient_cauchy.__func__,
        "arctan": gradient_arctan.__func__,
    }



##
# \file ParameterStudy.py
# \brief      Abstract class to define paths to predefined paths to write and
#             read for parameter studies
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       Aug 2017
#

import os
import itertools
from abc import ABCMeta, abstractmethod

import pysitk.python_helper as ph
from nsol.definitions import FILENAME_EXTENSION


##
# Abstract class holding the information where to store paths to read and write
# for parameter studies
# \date       2017-08-05 18:44:56+0100
#
class ParameterStudy(object):
    __metaclass__ = ABCMeta

    def __init__(self, directory, name):
        self._directory = directory
        self._name = name

    def get_parameter_study_name(self):
        return self._name

    ##
    # Gets the path to file where the parameter configurations are stored.
    # \date       2017-08-05 18:45:57+0100
    #
    # \param      self       The object
    # \param      specifier  Specifier of parameter file (for easier reading
    #                        later)
    #
    # \return     The path to file parameters.
    #
    def _get_path_to_file_parameters(self, specifier="_parameters"):
        return os.path.join(self._directory,
                            self._name + specifier + "." + FILENAME_EXTENSION)

    ##
    # Gets the path to file for each computed measure.
    # \date       2017-08-05 18:46:31+0100
    #
    # \param      self       The object
    # \param      measure    Measure ID as string
    # \param      specifier  Specifier of measure files (for easier reading
    #                        later)
    #
    # \return     The path to file measures.
    #
    def _get_path_to_file_measures(self, measure, specifier="_measure_"):
        return os.path.join(self._directory,
                            self._name + specifier + measure +
                            "." + FILENAME_EXTENSION)

    ##
    # Gets the path to file where computational times are stored
    # \date       2017-08-05 18:48:13+0100
    #
    # \param      self       The object
    # \param      specifier  Specifier of computational time file (for easier
    #                        reading later)
    #
    # \return     The path to file holding computational time.
    #
    def _get_path_to_file_computational_time(self,
                                             specifier="_computational_time"):
        return os.path.join(self._directory,
                            self._name + specifier + "." + FILENAME_EXTENSION)

    ##
    # Gets the path to file where reconstructions after n iterations are stored
    # \date       2017-08-05 18:48:13+0100
    #
    # \param      self       The object
    # \param      specifier  Specifier of reconstruction file (for easier
    #                        reading later)
    #
    # \return     The path to file holding reconstructions.
    #
    def _get_path_to_file_reconstructions(self,
                                          specifier="_reconstructions"):
        return os.path.join(self._directory,
                            self._name + specifier + ".npz")



# \file DataReader.py
#  \brief Reads data and returns numpy arrays
#
#  \author Michael Ebner (michael.ebner.14@ucl.ac.uk)
#  \date September 2017


# Import libraries
import os
import sys
import SimpleITK as sitk
import scipy.io
import numpy as np
import re
import natsort

import pysitk.python_helper as ph
import pysitk.simple_itk_helper as sitkh


class DataReader(object):

    def __init__(self, path_to_file):
        self._path_to_file = path_to_file

        # Get filename extension
        self._file_type = os.path.basename(self._path_to_file).split(".")[1]

        self._read_data = {
            "png": self._read_data_png,
            "mat": self._read_data_mat,
            "nii": self._read_data_nii,
        }

        self._nda = None
        self._image_sitk = None

    def read_data(self):
        if not ph.file_exists(self._path_to_file):
            raise IOError("Filename '%s' not found" % (self._path_to_file))

        self._read_data[self._file_type]()

    def get_data(self):
        return np.array(self._nda, dtype=np.float64)

    def get_image_sitk(self):
        return self._image_sitk

    def _read_data_png(self):
        self._nda = ph.read_image(self._path_to_file)

    def _read_data_mat(self):
        dic = scipy.io.loadmat(self._path_to_file)
        ndas = [dic[k] for k in dic.keys() if isinstance(dic[k], np.ndarray)]

        if len(ndas) > 1:
            raise IOError("MAT file '%s' must include one array only" %
                          (self._path_to_file))

        self._nda = ndas[0]

    def _read_data_nii(self):
        self._image_sitk = sitkh.read_nifti_image_sitk(
            self._path_to_file)
        self._nda = sitk.GetArrayFromImage(self._image_sitk)



##
# \file ParameterStudyReader.py
# \brief Read parameter studies obtained by ParameterStudy object
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       Aug 2017
#

import os
import numpy as np
import re
import natsort

import pysitk.python_helper as ph

from nsol.parameter_study import ParameterStudy
from nsol.definitions import REGEX_FILENAMES
from nsol.definitions import FILENAME_EXTENSION


class ReaderParameterStudy(ParameterStudy):

    def __init__(self, directory, name):
        self._directory = directory
        self._name = name

        # List of lines of read parameter file
        self._lines_params = None

    ##
    # Reads the specified parameter study.
    # \date       2017-08-05 19:54:59+0100
    #
    # \param      self  The object
    #
    def read_study(self):

        if not ph.directory_exists(self._directory):
            raise ValueError("Directory '%s' does not exist" %
                             (self._directory))

        # Get all measures of study
        self._measures = self._get_measure_names()

        if len(self._measures) == 0:
            raise RuntimeError("No measures to study '%s' found in '%s'"
                               % (self._name, self._directory))

        # Store the lines of the parameter file which hold the
        # information on the used parameters for further processing
        self._lines_params = self._read_parameter_file_lines()

        # Rebuild dictionary which was used for the parameter study as input
        self._parameters_dic = self._get_parameters()

    ##
    # Gets the reconstructions.
    # \date       2017-09-12 14:09:51+0100
    #
    # \param      self  The object
    #
    # \return     All reconstructions and associated information as dictionary.
    #
    def get_reconstructions(self):
        if not ph.file_exists(self._get_path_to_file_reconstructions()):
            raise IOError("File '%s' not available" %
                          self._get_path_to_file_reconstructions())
        return np.load(self._get_path_to_file_reconstructions())

    ##
    # Gets the names of all evaluated measures belonging to selected study.
    # \date       2017-08-05 19:53:46+0100
    #
    # \param      self  The object
    #
    # \return     Names of all evaluated measures as list of strings.
    #
    def get_measures(self):
        self._check_that_study_was_read()
        return self._measures

    ##
    # Gets the results for selected measure.
    #
    # Return results for selected measure as numpy data array for all parameter
    # runs (rows) and all iterations (columns).
    # \date       2017-08-05 19:32:12+0100
    #
    # \param      self     The object
    # \param      measure  ID of measure whose results shall be returned,
    #                      string
    #
    # \return     Results of measure as 2D numpy data array
    #
    def get_results(self, measure):
        return np.loadtxt(self._get_path_to_file_measures(measure), skiprows=2)

    ##
    # Gets the dictionary of the parameters which were used for the specified
    # parameter study as input.
    #
    # E.g. parameters = {
    #   "alpha": [0.02, 0.07, 0.12, 0.17, 0.22, 0.27, 0.32, 0.37, 0.42, 0.47],
    #   "data_loss": ["linear", "arctan"],
    #   "data_loss_scale": [1., 1.1],
    # }
    # \date       2017-08-05 19:44:30+0100
    #
    # \param      self  The object
    #
    # \return     The parameters as dictionary.
    #
    def get_parameters(self):
        self._check_that_study_was_read()
        return self._parameters_dic

    ##
    # Gets the dictionary linking the parameters with the line in the result
    # array.
    #
    # E.g. dict = {
    #   ('0.02', 'linear', '1.'): 0,
    #   ('0.02', 'linear', '1.1'): 1,
    #   ('0.02', 'arctan', '1.'): 2,
    #   ...
    # }
    # \date       2017-08-05 21:04:17+0100
    #
    # \param      self  The object
    #
    # \return     The parameters to line.
    #
    def get_parameters_to_line(self):
        self._check_that_study_was_read()

        # Build dictionary which links parameter set to line of result array
        return self._get_parameters_to_line()

    ##
    # Gets the lines to given parameters where only key of dictionary is
    # varying.
    #
    # E.g. parameters = {
    #   'alpha': [0.02, 0.07, 0.12, 0.17, 0.22, 0.27, 0.32, 0.37, 0.42, 0.47],
    #   'data_loss': 'arctan',
    #   'data_loss_scale': 1.,
    # }
    # The returned lines refer to the respective entry of the array obtained by
    # get_results()
    # \date       2017-08-05 22:08:20+0100
    #
    # \param      self        The object
    # \param      parameters  dictionary where only one parameter is varying
    #
    # \return     The lines to associated parameters as 1D numpy array.
    #
    def get_lines_to_parameters(self, parameters):
        self._check_that_study_was_read()
        return self._get_lines_to_parameters(parameters)

    ##
    # Gets dictionary which links each line/row data array results with
    # readable parameter configuration.
    #
    # E.g. dict = {
    #   0: 'alpha=0.02, rho=0.1',
    #   1: 'alpha=0.02, rho=0.5',
    #   2: 'alpha=0.02, rho=1.0',
    # }
    # \date       2017-08-05 19:45:28+0100
    #
    # \param      self       The object
    # \param      separator  Separator between parameter information, string
    # \param      compact    Turn on/off compact format, bool
    #
    # \return     Dictionary to link line/row of array with parameters.
    #
    def get_line_to_parameter_labels(self, separator=", ", compact=False):
        # Link line of numpy array results to readable label
        return self._get_line_to_parameter_labels(separator=separator,
                                                  compact=compact)

    def _read_parameter_file_lines(self):

        # Fetch information on used parameters from file. List of strings
        # holds information of a single line in file
        path_to_file_parameters = self._get_path_to_file_parameters()
        lines = ph.read_file_line_by_line(path_to_file_parameters)

        # Skip first line which contains header information only
        lines = lines[1:]

        # Remove comments and newline at the end of each line. Only relevant
        # parameter information remains
        for i in range(len(lines)):
            lines[i] = re.sub("## ", "", lines[i])
            lines[i] = re.sub("\n", "", lines[i])

        # Lines with information on the varying parameters of the study
        return lines

    def _get_measure_names(self):
        pattern = self._name + \
            "_measure_(" + REGEX_FILENAMES + ")[.]" + FILENAME_EXTENSION
        p = re.compile(pattern)

        measures = [p.match(f).group(1)
                    for f in os.listdir(self._directory) if p.match(f)]

        return measures

    def _get_parameters(self):

        lines = list(self._lines_params)

        # Link parameter name with parameter value
        parameters_dic = {}

        # split at tabs to get filenames and forget about first line
        parameters = lines[0].split("\t")
        lines = lines[1:]

        for i, parameter in enumerate(parameters):

            tmp = [line.split("\t")[i] for line in lines]
            tmp = list(set(tmp))

            tmp = natsort.natsorted(tmp, key=lambda y: y.lower())

            try:
                tmp = [float(t) for t in tmp]
                tmp.sort()

            except:
                pass

            parameters_dic[parameter] = tmp

        return parameters_dic

    def _get_parameters_to_line(self):

        # Skip parameter information header
        lines = self._lines_params[1:]

        parameters_to_line_dic = {}
        for i, line in enumerate(lines):
            parameters = [v for v in line.split("\t")]
            parameters_to_line_dic[tuple(parameters)] = i

        return parameters_to_line_dic

    def _get_lines_to_parameters(self, parameters):

        if parameters.keys() != self._parameters_dic.keys():
            raise ValueError("Provided dictionary keys must match. "
                             "Required keys for this study are " +
                             str(self._parameters_dic.keys()))

        # Find single varying key
        varying_key = None
        for key in parameters.keys():
            if type(parameters[key]) in (tuple, list, np.ndarray):
                if len(parameters[key]) == 1:
                    raise ValueError("Single entry in key '%s' must not "
                                     "be a list" % (key))

                elif len(parameters[key]) > 1:
                    if varying_key is None:
                        varying_key = key
                        rows = len(parameters[key])
                    else:
                        raise ValueError("Provided dictionary can only vary "
                                         "in a single key")

        parameters_to_line_dic = self.get_parameters_to_line()
        lines = np.zeros(rows, dtype=int)
        for i in range(rows):
            key = []
            for k in parameters.keys():
                if k == varying_key:
                    key.append(str(parameters[k][i]))
                else:
                    key.append(str(parameters[k]))

            line = parameters_to_line_dic[tuple(key)]
            lines[i] = line

        lines = sorted(lines)

        return lines

    def _get_line_to_parameter_labels(self, separator, compact):

        # Skip parameter information header
        lines = self._lines_params[1:]

        # Link parameter values to line in text files
        line_to_parameters_dic = {}
        for i, line in enumerate(lines):
            if compact:
                # e.g. dict[2] = '0.1,2,linear'
                label = (separator).join(line.split("\t"))
            else:
                # e.g. dict[2] = 'alpha=0.1, rho=2, data_loss=linear'
                label = [k + "=" + v for (k, v) in
                         zip(self._parameters_dic.keys(), line.split("\t"))]
                label = (separator).join(label)
            line_to_parameters_dic[i] = label

        return line_to_parameters_dic

    ##
    # Check that study was read to obtain information on parameters
    # \date       2017-08-05 22:44:12+0100
    #
    # \param      self  The object
    #
    def _check_that_study_was_read(self):
        if self._lines_params is None:
            raise UnboundLocalError("Execute 'read_study' first to get "
                                    "information on parameters.")



#!/usr/bin/python

##
# \file run_denoising.py
# \brief      Run TK0L2/TK1L2/TVL2/HuberL2 deconvolution
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       Sept 2017
#

import os
import sys
import scipy.io
import numpy as np
import SimpleITK as sitk

import pysitk.python_helper as ph
import pysitk.simple_itk_helper as sitkh

import nsol.linear_operators as LinearOperators
import nsol.data_reader as dr
import nsol.data_writer as dw
import nsol.observer as Observer
import nsol.input_argparser as InputArgparser
import nsol.deconvolution_solver_parameter_study_interface as interface


def main():

    input_parser = InputArgparser.InputArgparser(
        description="Run TK0L2/TK1L2/TVL2/HuberL2 deconvolution",
    )
    input_parser.add_observation(required=True)
    input_parser.add_result(required=False)
    input_parser.add_reference(required=False)
    input_parser.add_blur(default=1)
    input_parser.add_reconstruction_type(default="TVL2")
    input_parser.add_measures(default=["PSNR", "RMSE", "SSIM", "NCC", "NMI"])
    input_parser.add_iterations(default=50)
    input_parser.add_solver(default="PD")
    input_parser.add_rho(default=0.1)
    input_parser.add_alpha(default=[0.01])
    input_parser.add_data_loss(default="linear")
    input_parser.add_data_loss_scale(default=1)
    input_parser.add_iter_max(default=10)
    input_parser.add_minimizer(default="lsmr")
    input_parser.add_dir_output_figures(default=None)
    input_parser.add_verbose(default=0)
    args = input_parser.parse_args()
    input_parser.print_arguments(args)

    observers = [None] * len(args.alpha)
    recons = [None] * len(args.alpha)
    data_nda = []
    data_labels = []

    if len(args.alpha) > 1 and args.result is not None:
        ph.print_warning("Multiple alphas overwrite result")
    elif len(args.alpha) == 1 and args.result is None:
        raise IOError("'--result' must be specified")

    if args.result is None:
        ph.print_warning("No output ('--result') provided")

    # --------------------------------Read Data--------------------------------
    data_reader = dr.DataReader(args.observation)
    data_reader.read_data()
    observed_nda = data_reader.get_data()
    dimension = observed_nda.ndim

    data_nda.append(observed_nda)
    if dimension < 3:
        data_labels.append("observed:\n%s" %
                           os.path.basename(args.observation))
    elif dimension == 3:
        data_labels.append(
            os.path.basename(args.observation).split(".")[0])

    if args.reference is not None:
        data_reader = dr.DataReader(args.reference)
        data_reader.read_data()
        reference_nda = data_reader.get_data()
        data_nda.append(reference_nda)
        if dimension < 3:
            data_labels.append("reference:\n%s" %
                               os.path.basename(args.reference))
        elif dimension == 3:
            data_labels.append(
                os.path.basename(args.reference).split(".")[0])
        x_ref = reference_nda.flatten()
    else:
        x_ref = None

    if args.blur > 0:
        sigma = np.atleast_1d(args.blur)
        if sigma.ndim != observed_nda.ndim:
            try:
                cov = np.diag(np.ones(observed_nda.ndim)) * sigma**2
            except:
                raise IOError(
                    "Blur information must be either 1- or d-dimensional")

    # ------------------------------Set Up Solver------------------------------
    b = observed_nda.flatten()
    x0 = observed_nda.flatten()
    x_scale = np.max(observed_nda)

    # Get spacing for blurring operator
    if data_reader.get_image_sitk() is None:
        spacing = np.ones(observed_nda.ndim)
    else:
        spacing = np.array(data_reader.get_image_sitk().GetSpacing())

    # Get linear operators
    linear_operators = eval(
        "LinearOperators.LinearOperators%dD" % dimension)(spacing=spacing)
    A, A_adj = linear_operators.get_gaussian_blurring_operators(cov)
    grad, grad_adj = linear_operators.get_gradient_operators()

    # A: X \rightarrow Y and D: X \rightarrow Z
    X_shape = observed_nda.shape
    Y_shape = A(observed_nda).shape
    Z_shape = grad(observed_nda).shape

    A_1D = lambda x: A(x.reshape(*X_shape)).flatten()
    A_adj_1D = lambda x: A_adj(x.reshape(*Y_shape)).flatten()

    D_1D = lambda x: grad(x.reshape(*X_shape)).flatten()
    D_adj_1D = lambda x: grad_adj(x.reshape(*Z_shape)).flatten()

    solver_interface = interface.DeconvolutionSolverStudyInterface(
        A=A_1D,
        A_adj=A_adj_1D,
        D=D_1D,
        D_adj=D_adj_1D,
        b=b,
        x0=x0,
        alpha=args.alpha[0],
        x_scale=x_scale,
        data_loss=args.data_loss,
        data_loss_scale=args.data_loss_scale,
        iter_max=args.iter_max,
        iterations=args.iterations,
        minimizer=args.minimizer,
        measures=args.measures,
        dimension=dimension,
        reconstruction_type=args.reconstruction_type,
        rho=args.rho,
        x_ref=x_ref,
        tv_solver=args.solver,
        verbose=args.verbose,
    )
    solver_interface.set_up_solver()
    solver_interface.set_up_measures()

    solver = solver_interface.get_solver()
    measures_dic = solver_interface.get_measures()

    # -----------------------------Run Reconstruction--------------------------
    for i, alpha in enumerate(args.alpha):
        ph.print_subtitle("Iteration %d/%d" % (i+1, len(args.alpha)))
        if dimension < 3:
            title_prefix = args.reconstruction_type + \
                " (" + r"$\alpha=%g$)" % alpha
        elif dimension == 3:
            title_prefix = ("%s_alpha=%g" % (
                args.reconstruction_type, alpha)).replace(".", "p")
        solver.set_alpha(alpha)

        observers[i] = Observer.Observer()
        observers[i].set_measures(measures_dic)
        solver.set_observer(observers[i])

        solver.run()
        recons[i] = np.array(solver.get_x().reshape(*X_shape))
        print("\nComputational time %s: %s" %
              (args.reconstruction_type, solver.get_computational_time()))

        data_nda.append(recons[i])
        data_labels.append(title_prefix)

        if args.result is not None:
            data_writer = dw.DataWriter(
                recons[i], args.result, data_reader.get_image_sitk())
            data_writer.write_data()

    # ----------------------------Visualize Results----------------------------
    if args.verbose:
        filename_prefix = args.reconstruction_type

        if dimension == 2:
            ph.show_arrays(
                data_nda,
                title=data_labels,
                fig_number=None,
                # cmap="jet",
                cmap="Greys_r",
                use_same_scaling=True,
                directory=args.dir_output_figures,
                filename=args.reconstruction_type+"_comparison.pdf",
                save_figure=0 if args.dir_output_figures is None else 1,
                fontsize=8,
            )

        elif dimension == 3:
            images_sitk = [None] * len(data_nda)
            observation_sitk = data_reader.get_image_sitk()
            for i, nda in enumerate(data_nda):
                images_sitk[i] = sitk.GetImageFromArray(nda)
                images_sitk[i].CopyInformation(observation_sitk)
            sitkh.show_sitk_image(images_sitk, label=data_labels)

        # if args.reference is not None:
        linestyles = ["-", ":", "-", "-."] * 10

        if args.reference is not None:
            linestyles = ["-", ":", "-", "-."] * 10

            for ob in observers:
                ob.compute_measures()

            for k in measures_dic.keys():
                x = []
                y = []
                legend = []
                for i, alpha in enumerate(args.alpha):
                    title = args.reconstruction_type + ": %s" % k
                    y.append(observers[i].get_measures()[k])
                    legend.append(r"$\alpha=%g$" % alpha)
                    x.append(range(0, len(y[-1])))
                ph.show_curves(
                    y=y,
                    x=x,
                    xlabel="iteration",
                    labels=legend,
                    title=title,
                    linestyle=linestyles,
                    markers=ph.MARKERS,
                    markevery=10,
                    directory=args.dir_output_figures,
                    filename=filename_prefix+"_"+k+".pdf",
                    save_figure=0 if args.dir_output_figures is None else 1,
                )

    return 0

if __name__ == '__main__':
    main()



##
# \file linear_solver.py
# \brief      Abstract class to define a numerical solver for linear
#             least-squares problems
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       July 2017
#

import os
import sys
import numpy as np
from abc import ABCMeta, abstractmethod

import pysitk.python_helper as ph

from nsol.solver import Solver
from nsol.loss_functions import LossFunctions as lf


##
# Abstract class to define a numerical solver for linear least-squares
# problems.
#
# Abstract class to define a numerical solver for linear least-squares problems
# with (non-linear) data loss function rho and regularizer g, i.e.
# max_x [1/2 ||rho( Ax-b )||^2 + alpha g(x)]
# \date       2017-07-20 23:24:24+0100
#
class LinearSolver(Solver):
    __metaclass__ = ABCMeta

    ##
    # Store relevant information for linear solvers
    # \date       2017-07-20 23:27:49+0100
    #
    # \param      self             The object
    # \param      A                Function associated to linear operator A:
    #                              X->Y; x->A(x) with x being a 1D numpy array
    # \param      A_adj            Function associated to adjoint linear
    #                              operator A^*: Y->X; y->A^*(y)
    # \param      b                Right hand-side of linear system Ax = b as
    #                              1D numpy array
    # \param      x0               Initial value as 1D numpy array
    # \param      alpha            Regularization parameter; scalar
    # \param      x_scale          Characteristic scale of each variable.
    #                              Setting x_scale is equivalent to
    #                              reformulating the problem in scaled
    #                              variables ``xs = x / x_scale``
    # \param      data_loss        Data loss function rho specified as string,
    #                              e.g. "linear", "soft_l1", "huber", "cauchy",
    #                              "arctan".
    # \param      data_loss_scale  Value of soft margin between inlier and
    #                              outlier residuals, default is 1.0. The loss
    #                              function is evaluated as rho_(f2) = C**2 *
    #                              rho(f2 / C**2), where C is data_loss_scale.
    #                              This parameter has no effect with
    #                              data_loss='linear', but for other loss
    #                              values it is of crucial importance.
    # \param      minimizer        String defining the used optimizer, i.e.
    #                              "lsmr", "least_squares" or any solver as
    #                              provided by scipy.optimize.minimize
    # \param      iter_max         Number of maximum iterations for used
    #                              minimizer, integer value
    # \param      verbose          Verbose output, bool
    #
    def __init__(self,
                 A,
                 A_adj,
                 b,
                 x0,
                 alpha,
                 x_scale,
                 data_loss,
                 data_loss_scale,
                 minimizer,
                 iter_max,
                 verbose):

        Solver.__init__(self, x0=x0, x_scale=x_scale, verbose=verbose)

        self._A = A
        self._A_adj = A_adj
        self._b = b / self._x_scale
        self._alpha = float(alpha)
        self._data_loss = data_loss
        self._data_loss_scale = float(data_loss_scale)
        self._minimizer = minimizer
        self._iter_max = iter_max

    ##
    # Gets the linear operator A
    # \date       2017-09-06 16:29:11+0100
    #
    # \param      self  The object
    #
    # \return     Forward operator as function: 1D -> 1D
    #
    def get_A(self):
        return self._A

    ##
    # Gets the adjoint linear operator A
    # \date       2017-09-06 16:29:11+0100
    #
    # \param      self  The object
    #
    # \return     Adjoint forward operator as function: 1D -> 1D
    #
    def get_A_adj(self):
        return self._A_adj

    ##
    # Gets the right hand-side b of linear system Ax = b
    # \date       2017-09-06 16:37:32+0100
    #
    # \param      self  The object
    #
    # \return     Right hand-side as 1D array
    #
    def get_b(self):
        return np.array(self._b) * self._x_scale

    ##
    # Sets the regularization parameter alpha.
    # \date       2017-08-04 18:54:24+0100
    #
    # \param      self   The object
    # \param      alpha  Regularization  parameter; scalar
    #
    def set_alpha(self, alpha):
        self._alpha = alpha

    ##
    # Gets the regularization parameter alpha.
    # \date       2017-08-04 18:54:59+0100
    #
    # \param      self  The object
    #
    # \return     scalar
    #
    def get_alpha(self):
        return self._alpha

    ##
    # Sets the data loss function specified as string
    # \date       2017-08-04 18:55:38+0100
    #
    # \param      self       The object
    # \param      data_loss  The data loss as defined in LossFunctions, string
    #
    def set_data_loss(self, data_loss):
        if data_loss not in lf.get_loss.keys():
            raise ValueError("data_loss must be in " +
                             str(lf.get_loss.keys()))
        self._data_loss = data_loss

    ##
    # Gets the data loss.
    # \date       2017-08-04 19:00:42+0100
    #
    # \param      self  The object
    #
    # \return     The data loss function specifier, string
    #
    def get_data_loss(self):
        return self._data_loss

    ##
    # Sets the data loss scale.
    # \date       2017-08-04 19:00:54+0100
    #
    # \param      self             The object
    # \param      data_loss_scale  Value of soft margin between inlier and
    #                              outlier residuals, default is 1.0. The loss
    #                              function is evaluated as rho_(f2) = C**2 *
    #                              rho(f2 / C**2), where C is data_loss_scale.
    #                              This parameter has no effect with
    #                              data_loss='linear', but for other loss
    #                              values it is of crucial importance.
    #
    def set_data_loss_scale(self, data_loss_scale):
        self._data_loss_scale = data_loss_scale

    ##
    # Gets the data loss scale.
    # \date       2017-08-04 19:01:57+0100
    #
    # \param      self  The object
    #
    # \return     The data loss scale; scalar
    #
    def get_data_loss_scale(self):
        return self._data_loss_scale

    ##
    # Sets the minimizer.
    # \date       2017-08-04 19:06:16+0100
    #
    # \param      self       The object
    # \param      minimizer  String defining the used optimizer, i.e. "lsmr",
    #                        "least_squares" or any solver as provided by
    #                        scipy.optimize.minimize
    #
    def set_minimizer(self, minimizer):
        self._minimizer = minimizer

    ##
    # Gets the minimizer.
    # \date       2017-08-04 19:06:58+0100
    #
    # \param      self  The object
    #
    # \return     The minimizer as string
    #
    def get_minimizer(self):
        return self._minimizer

    ##
    # Sets the number of maximum iterations.
    # \date       2017-08-04 19:07:23+0100
    #
    # \param      self      The object
    # \param      iter_max  Number of maximum iterations for used minimizer,
    #                       integer value
    #
    def set_iter_max(self, iter_max):
        self._iter_max = iter_max

    ##
    # Gets the iterator maximum.
    # \date       2017-08-04 19:08:06+0100
    #
    # \param      self  The object
    #
    # \return     Number of maximum iterations for used minimizer, integer
    #             value
    #
    def get_iter_max(self):
        return self._iter_max

    ##
    # Gets the total cost as 1/2 ||rho( Ax-b )||^2 + alpha g(x)
    # \date       2017-07-20 23:38:46+0100
    #
    # \param      self  The object
    #
    # \return     The total cost as scalar value
    #
    def get_total_cost(self):
        data_cost = self.get_cost_data_term()
        regularization_cost = self.get_cost_regularization_term()
        return data_cost + self._alpha * regularization_cost

    ##
    # Gets the cost of the data term, i.e. f(x) = 1/2 ||rho( Ax-b )||^2
    # \date       2017-07-20 23:56:30+0100
    #
    # \param      self  The object
    #
    # \return     The cost of the data term.
    #
    def get_cost_data_term(self):
        return self._get_cost_data_term(self._x)

    ##
    # Gets the ell2 cost of the data term, i.e.  1/2 || Ax-b ||^2
    # \date       2017-07-20 23:58:34+0100
    #
    # \param      self  The object
    #
    # \return     The ell2 cost data term.
    #
    def get_ell2_cost_data_term(self):
        return self._get_ell2_cost_data_term(self._x)

    ##
    # Gets the cost of the regularization term g(x)
    # \date       2017-07-21 00:00:26+0100
    #
    # \param      self  The object
    #
    # \return     The cost of the regularization term.
    #
    def get_cost_regularization_term(self):
        return self._get_cost_regularization_term(self._x)

    ##
    # Prints the statistics of the performed optimization
    # \date       2017-07-21 00:01:10+0100
    #
    # \param      self  The object
    # \param      fmt   Format for printing numerical values
    #
    def print_statistics(self, fmt="%.3e"):

        cost_data = self.get_cost_data_term()
        cost_data_ell2 = self.get_ell2_cost_data_term()
        cost_regularizer = self.get_cost_regularization_term()

        ph.print_subtitle("Summary Optimization")
        ph.print_info("Computational time: %s" %
                      (self.get_computational_time()))
        ph.print_info("Cost data term (f, loss=%s, scale=%g): " %
                      (self._data_loss, self._data_loss_scale) +
                      fmt % (cost_data) +
                      " (ell2-cost: " + fmt % (cost_data_ell2) + ")")
        ph.print_info(
            "Cost regularization term (g): " +
            fmt % (cost_regularizer))
        ph.print_info(
            "Total cost (f + alpha g; alpha = %g" % (self._alpha) + "): " +
            fmt % (cost_data + self._alpha * cost_regularizer))

    def _get_cost_data_term(self, x):

        residual = self._A(x) - self._b
        cost = 0.5 * np.sum(
            lf.get_loss[self._data_loss](f2=residual ** 2,
                                         f_scale=self._data_loss_scale))

        return cost

    def _get_ell2_cost_data_term(self, x):

        residual = self._A(x) - self._b
        cost = 0.5 * np.sum(residual ** 2)

        return cost

    def _get_gradient_cost_data_term(self, x):

        residual = self._A(x) - self._b

        grad = self._A_adj(
            lf.get_gradient_loss[self._data_loss](f2=residual ** 2,
                                                  f_scale=self._data_loss_scale
                                                  ) * residual)

        return grad

    @abstractmethod
    def _get_cost_regularization_term(self, x):
        pass



#!/usr/bin/python

##
# \file run_tests.py
# \brief      main-file to run specified unit tests
#
# \author     Michael Ebner (michael.ebner.14@ucl.ac.uk)
# \date       July 2017
#


import unittest

# Import modules for unit testing
from kernels_test import *
from loss_functions_test import *
from similarity_measures_test import *
from solvers_test import *
from run_denoising_test import *

if __name__ == '__main__':
    print("\nUnit tests:\n--------------")
    unittest.main()



from nsol.__about__ import (
    __author__,
    __email__,
    __license__,
    __summary__,
    __title__,
    __uri__,
    __version__,
 )

__all__ = [
    "__author__",
    "__email__",
    "__license__",
    "__summary__",
    "__title__",
    "__uri__",
    "__version__",
]
